@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python -i
 
 # Fabric - Pythonic remote deployment tool.
 # Copyright (C) 2008  Christian Vest Hansen
@@ -29,23 +29,14 @@ import time
 import types
 import datetime
 
-import_error = False
 try:
     import paramiko as ssh
 except ImportError:
     print("Error: paramiko is a required module. Please install it:")
     print("  $ sudo easy_install paramiko")
-    import_error = True
-try:
-    from decorator import decorator
-except ImportError:
-    print("Error: decorator is a required module. Please install it:")
-    print("  $ sudo easy_install decorator")
-    import_error = True
-if import_error:
-    exit(1)
+    sys.exit(1)
 
-__version__ = '0.0.2'
+__version__ = '0.0.5'
 __author__ = 'Christian Vest Hansen'
 __author_email__ = 'karmazilla@gmail.com'
 __url__ = 'https://savannah.nongnu.org/projects/fab/'
@@ -60,26 +51,26 @@ __greeter__ = '''\
 ENV = {
     'fab_version':__version__,
     'fab_author':__author__,
-    'fab_mode':'fanout',
+    'fab_mode':'rolling',
     'fab_port':22,
-    'fab_user':None,
+    'fab_user':os.getlogin(),
     'fab_pkey':None,
     'fab_key_filename':None,
     'fab_new_host_key':'accept',
     'fab_shell':'/bin/bash -l -c "%s"',
-    # TODO: make fab_timestamp UTC
-    'fab_timestamp':datetime.datetime.now().strftime('%F_%H-%M-%S'),
-    'fab_debug':False,
+    'fab_timestamp':datetime.datetime.utcnow().strftime('%F_%H-%M-%S'),
+    'fab_print_real_sudo':False,
 }
-COMMANDS = {}
+
 CONNECTIONS = []
+COMMANDS = {}
 OPERATIONS = {}
+STRATEGIES = {}
 _LAZY_FORMAT_SUBSTITUTER = re.compile(r'\$\((?P<var>\w+?)\)')
 
 #
-# Helper decorators
+# Helper decorators:
 #
-
 def new_registering_decorator(registry):
     def registering_decorator(first_arg=None):
         if callable(first_arg):
@@ -93,14 +84,24 @@ def new_registering_decorator(registry):
     return registering_decorator
 command = new_registering_decorator(COMMANDS)
 operation = new_registering_decorator(OPERATIONS)
+strategy = new_registering_decorator(STRATEGIES)
 
+def run_per_host(op_fn):
+    def wrapper(*args, **kvargs):
+        if not CONNECTIONS:
+            _connect()
+        _on_hosts_do(op_fn, *args, **kvargs)
+    wrapper.__doc__ = op_fn.__doc__
+    wrapper.__name__ = op_fn.__name__
+    return wrapper
 
 #
 # Standard fabfile operations:
 #
 @operation
 def set(**variables):
-    """Set a number of Fabric environment variables.
+    """
+    Set a number of Fabric environment variables.
     
     Set takes a number of keyword arguments, and defines or updates the
     variables that corrosponds to each keyword with the respective value.
@@ -122,24 +123,27 @@ def set(**variables):
             ENV[k] = v
 
 @operation
-def get(name):
-    """Get the value of a given Fabric environment variable.
+def get(name, otherwise=None):
+    """
+    Get the value of a given Fabric environment variable.
     
-    If the variable isn't found, then this operation returns None.
+    If the variable isn't found, then this operation returns the
+    value of the 'otherwise' parameter, which is None unless set.
     
     """
-    return name in ENV and ENV[name] or None
+    return name in ENV and ENV[name] or otherwise
 
 @operation
 def require(var, **kvargs):
-    """Make sure that a certain environmet variable is available.
+    """
+    Make sure that a certain environmet variable is available.
     
     The 'var' parameter is a string that names the variable to check for.
     Two other optional kvargs are supported:
-     - 'used_for' is a string that gets injected into, and then printed, as
-       something like this string: "This variable is used for %s".
-     - 'provided_by' is a list of strings that name commands which the user
-       can run in order to satisfy the requirement.
+        * 'used_for' is a string that gets injected into, and then printed, as
+          something like this string: "This variable is used for %s".
+        * 'provided_by' is a list of strings that name commands which the user
+          can run in order to satisfy the requirement.
     
     If the required variable is not found in the current environment, then the
     operation is stopped and Fabric halts.
@@ -158,15 +162,62 @@ def require(var, **kvargs):
         + "' variable.") % ENV
     )
     if 'used_for' in kvargs:
-        print("This variable is used for %s" % kvargs['used_for'])
+        print("This variable is used for %s" % _lazy_format(
+            kvargs['used_for']))
     if 'provided_by' in kvargs:
         print("Get the variable by running one of these commands:")
-        print('\t' + ('\n\t'.join(kvargs['provided_by'])))
-    exit(1)
+        to_s = lambda obj: getattr(obj, '__name__', str(obj))
+        provided_by = [to_s(obj) for obj in kvargs['provided_by']]
+        print('\t' + ('\n\t'.join(provided_by)))
+    sys.exit(1)
 
 @operation
-def put(localpath, remotepath, **kvargs):
-    """Upload a file to the current hosts.
+def prompt(varname, msg, validate=None, default=None):
+    """
+    Display a prompt to the user and store the input in the given variable.
+    If the variable already exists, then it is not prompted for again.
+    
+    The 'validate' parameter is a callable that raises an exception on invalid
+    inputs and returns the input for storage in ENV.
+    It may process the input and convert it to a different type, as in the
+    second example below.
+    
+    Example:
+        # Simplest form:
+        prompt('environment', 'Please specify target environment')
+        
+        # With default:
+        prompt('dish', 'Specify favorite dish', default='spam & eggs')
+        
+        # With validation, i.e. require integer input:
+        prompt('nice', 'Please specify process nice level', validate=int)
+    
+    """
+    if varname in ENV and ENV[varname] is not None:
+        return
+    
+    if callable(default):
+        default = default()
+    
+    try:
+        default_str = default and (" [%s]" % str(default).strip()) or ""
+        prompt_msg = _lazy_format("%s%s: " % (msg.strip(), default_str))
+        value = raw_input(prompt_msg)
+        if not value:
+            value = default
+        
+        if callable(validate):
+            value = validate(value)
+        
+        set(**{varname: value})
+    except EOFError:
+        return
+
+@operation
+@run_per_host
+def put(host, client, env, localpath, remotepath, **kvargs):
+    """
+    Upload a file to the current hosts.
     
     The 'localpath' parameter is the relative or absolute path to the file on
     your localhost that you wish to upload to the fab_hosts.
@@ -183,13 +234,20 @@ def put(localpath, remotepath, **kvargs):
         put('bin/project.zip', '/tmp/project.zip')
     
     """
-    if not CONNECTIONS:
-        _connect()
-    _on_hosts_do(_put, localpath, remotepath, **kvargs)
+    localpath = _lazy_format(localpath, env)
+    remotepath = _lazy_format(remotepath, env)
+    if not os.path.exists(localpath):
+        return False
+    ftp = client.open_sftp()
+    print("[%s] put: %s -> %s" % (host, localpath, remotepath))
+    ftp.put(localpath, remotepath)
+    return True
 
 @operation
-def download(remotepath, localpath, **kvargs):
-    """Download a file from the remote hosts.
+@run_per_host
+def download(host, client, env, remotepath, localpath, **kvargs):
+    """
+    Download a file from the remote hosts.
     
     The 'remotepath' parameter is the relative or absolute path to the files
     to download from the fab_hosts. The 'localpath' parameter will be suffixed
@@ -210,13 +268,18 @@ def download(remotepath, localpath, **kvargs):
     respectively.
     
     """
-    if not CONNECTIONS:
-        _connect()
-    _on_hosts_do(_download, remotepath, localpath, **kvargs)
+    ftp = client.open_sftp()
+    localpath = _lazy_format(localpath) + '.' + host
+    remotepath = _lazy_format(remotepath)
+    print("[%s] download: %s <- %s" % (host, localpath, remotepath))
+    ftp.get(remotepath, localpath)
+    return True
 
 @operation
-def run(cmd, **kvargs):
-    """Run a shell command on the current fab_hosts.
+@run_per_host
+def run(host, client, env, cmd, **kvargs):
+    """
+    Run a shell command on the current fab_hosts.
     
     The provided command is executed with the permisions of fab_user, and the
     exact execution environ is determined by the fab_shell variable.
@@ -230,13 +293,22 @@ def run(cmd, **kvargs):
         run("ls")
     
     """
-    if not CONNECTIONS:
-        _connect()
-    _on_hosts_do(_run, cmd, **kvargs)
+    cmd = env['fab_shell'] % _lazy_format(cmd, env).replace('"', '\\"')
+    if not _confirm_proceed('run', host, kvargs):
+        return False # TODO: should we return False in fail??
+    print("[%s] run: %s" % (host, cmd))
+    stdin, stdout, stderr = client.exec_command(cmd)
+    out_th = _start_outputter("[%s] out" % host, stdout)
+    err_th = _start_outputter("[%s] err" % host, stderr)
+    out_th.join()
+    err_th.join()
+    return True
 
 @operation
-def sudo(cmd, **kvargs):
-    """Run a sudo (root privileged) command on the current hosts.
+@run_per_host
+def sudo(host, client, env, cmd, **kvargs):
+    """
+    Run a sudo (root privileged) command on the current hosts.
     
     The provided command is executed with root permisions, provided that
     fab_user is in the sudoers file in the remote host. The exact execution
@@ -252,13 +324,26 @@ def sudo(cmd, **kvargs):
         sudo("install_script.py")
     
     """
-    if not CONNECTIONS:
-        _connect()
-    _on_hosts_do(_sudo, cmd, **kvargs)
+    cmd = _lazy_format(cmd, env)
+    real_cmd = env['fab_shell'] % ("sudo -S " + cmd.replace('"', '\\"'))
+    cmd = env['fab_print_real_sudo'] and real_cmd or cmd
+    if not _confirm_proceed('sudo', host, kvargs):
+        return False # TODO: should we return False in fail??
+    print("[%s] sudo: %s" % (host, cmd))
+    stdin, stdout, stderr = client.exec_command(real_cmd)
+    stdin.write(env['fab_password'])
+    stdin.write('\n')
+    stdin.flush()
+    out_th = _start_outputter("[%s] out" % host, stdout)
+    err_th = _start_outputter("[%s] err" % host, stderr)
+    out_th.join()
+    err_th.join()
+    return True
 
 @operation
 def local(cmd, **kvargs):
-    """Run a command locally.
+    """
+    Run a command locally.
     
     This operation is essentially 'os.system()' except that variables are
     expanded prior to running.
@@ -281,11 +366,12 @@ def local(cmd, **kvargs):
             print("Warning: failed to execute command:")
             print("\t" + final_cmd)
         if failcode > 2:
-            exit(1)
+            sys.exit(1)
 
 @operation
 def local_per_host(cmd, **kvargs):
-    """Run a command locally, for every defined host.
+    """
+    Run a command locally, for every defined host.
     
     Like the local() operation, this is pretty similar to 'os.system()', but
     with this operation, the command is executed (and have its variables
@@ -304,7 +390,8 @@ def local_per_host(cmd, **kvargs):
 
 @operation
 def load(filename, **kvargs):
-    """Load up the given fabfile.
+    """
+    Load up the given fabfile.
     
     This loads the fabfile specified by the 'filename' parameter into fabric
     and make its commands and other functions available in the scope of the 
@@ -331,11 +418,12 @@ def load(filename, **kvargs):
         print("Warning: Cannot load file '%s'." % filename)
         print("No such file in your current directory.")
         if failcode > 2:
-            exit(1)
+            sys.exit(1)
 
 @operation
 def upload_project(**kvargs):
-    """Uploads the current project directory to the connected hosts.
+    """
+    Uploads the current project directory to the connected hosts.
     
     This is a higher-level convenience operation that basically 'tar' up the
     directory that contains your fabfile (presumably it is your project
@@ -362,7 +450,8 @@ def upload_project(**kvargs):
 #
 @command("help")
 def _help(**kvargs):
-    """Display usage help message to the console, or help for a given command.
+    """
+    Display Fabric usage help, or help for a given command.
     
     You can provide help with a parameter and get more detailed help for a
     specific command. For instance, to learn more about the list command, you
@@ -374,6 +463,9 @@ def _help(**kvargs):
     more about. For instance, to learn more about the 'run' operation, you
     could run 'fab help:op=run'.
     
+    Lastly, you can also learn more about a certain strategy with the 'strg'
+    and 'strategy' parameters: 'fab help:strg=rolling'.
+    
     """
     if kvargs:
         for k, v in kvargs.items():
@@ -383,10 +475,13 @@ def _help(**kvargs):
                 _print_help_for_in(k, OPERATIONS)
             elif k in ['op', 'operation']:
                 _print_help_for_in(kvargs[k], OPERATIONS)
+            elif k in ['strg', 'strategy']:
+                _print_help_for_in(kvargs[k], STRATEGIES)
             else:
                 _print_help_for(k, None)
     else:
-        print("""Fabric is a simple pythonic remote deployment tool.
+        print("""
+    Fabric is a simple pythonic remote deployment tool.
     
     Type 'fab list' to get a list of available commands.
     Type 'fab help:help' to get more information on how to use the built in
@@ -396,12 +491,14 @@ def _help(**kvargs):
 
 @command("list")
 def _list_commands(**kvargs):
-    """Display a list of commands with descriptions.
+    """
+    Display a list of commands with descriptions.
     
     By default, the list command prints a list of available commands, with a
     short description (if one is available). However, the list command can also
     print a list of available operaions if you provide it with the 'ops' or
-    'operations' parameter: 'fab list:ops'.
+    'operations' parameters, or it can print strategies with the 'strgs' and
+    'strategies' parameters.
     
     """
     if kvargs:
@@ -412,14 +509,18 @@ def _list_commands(**kvargs):
             elif k in ['ops', 'operations']:
                 print("Available operations are:")
                 _list_objs(OPERATIONS)
+            elif k in ['strgs', 'strategies']:
+                print("Available strategies are:")
+                _list_objs(STRATEGIES)
             else:
                 print("Don't know how to list '%s'." % k)
                 print("Try one of these instead:")
                 print(_indent('\n'.join([
                     'cmds', 'commands',
                     'ops', 'operations',
+                    'strgs', 'strategies',
                 ])))
-                exit(1)
+                sys.exit(1)
     else:
         print("Available commands are:")
         _list_objs(COMMANDS)
@@ -427,7 +528,7 @@ def _list_commands(**kvargs):
 @command("license")
 def _license():
     "Display the Fabric distribution license text."
-    print """		    GNU GENERAL PUBLIC LICENSE
+    print("""		    GNU GENERAL PUBLIC LICENSE
    TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 
   0. This License applies to any program or other work which contains
@@ -624,12 +725,12 @@ to ask for permission.  For software which is copyrighted by the Free
 Software Foundation, write to the Free Software Foundation; we sometimes
 make exceptions for this.  Our decision will be guided by the two goals
 of preserving the free status of all derivatives of our free software and
-of promoting the sharing and reuse of software generally."""
+of promoting the sharing and reuse of software generally.""")
 
 @command("warranty")
 def _warranty():
     "Display warranty information for the Fabric software."
-    print """			    NO WARRANTY
+    print("""			    NO WARRANTY
 
   BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
@@ -649,11 +750,12 @@ OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
 TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
 YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
 PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
-POSSIBILITY OF SUCH DAMAGES."""
+POSSIBILITY OF SUCH DAMAGES.""")
 
 @command("set")
 def _set(**kvargs):
-    """Set a Fabric variable.
+    """
+    Set a Fabric variable.
     
     Example:
         $fab set:fab_user=billy,other_var=other_value
@@ -679,6 +781,42 @@ def _shell(**kvargs):
         else:
             run(line)
 
+#
+# Standard strategies:
+#
+@strategy("fanout")
+def _fanout_strategy(fn, *args, **kvargs):
+    """
+    A strategy that executes on all hosts in parallel.
+    
+    THIS STRATEGY IS CURRENTLY BROKEN!
+    
+    """
+    threads = []
+    success = True
+    for host, client in CONNECTIONS:
+        env = dict(ENV)
+        env['fab_host'] = host
+        def functor():
+            success = success and fn(host, client, env, *args, **kvargs)
+        thread = threading.Thread(None, functor)
+        thread.setDaemon(True)
+        thread.start()
+        threads.append(thread)
+    for thread in threads:
+        thread.join()
+    return success
+
+@strategy("rolling")
+def _rolling_strategy(fn, *args, **kvargs):
+    """One-at-a-time fail-fast strategy."""
+    for host, client in CONNECTIONS:
+        env = dict(ENV)
+        env['fab_host'] = host
+        if not fn(host, client, env, *args, **kvargs):
+            return False
+    return True
+
 #
 # Internal plumbing:
 #
@@ -705,7 +843,7 @@ def _list_objs(objs):
     for name, fn in cmds:
         print '  ', name.ljust(max_name_len),
         if fn.__doc__:
-            print ':', fn.__doc__.splitlines()[0]
+            print ':', filter(None, fn.__doc__.splitlines())[0].strip()
         else:
             print
 
@@ -715,36 +853,38 @@ def _check_fab_hosts():
         print("Fabric requires a fab_hosts variable.")
         print("Please set it in your fabfile.")
         print("Example: set(fab_hosts=['node1.com', 'node2.com'])")
-        exit(1)
+        sys.exit(1)
 
 def _connect():
     "Populate CONNECTIONS with (hostname, client) tuples as per fab_hosts."
     _check_fab_hosts()
-    signal.signal(signal.SIGINT, lambda: _disconnect() and exit(0))
+    signal.signal(signal.SIGINT, lambda: _disconnect() and sys.exit(0))
     if 'fab_password' not in ENV:
         print(_lazy_format("Logging into the following hosts as $(fab_user):"))
         print(_indent('\n'.join(ENV['fab_hosts'])))
         ENV['fab_password'] = getpass.getpass()
     else:
         print("Warning: Putting your password in a fabfile is a bad idea.")
-    port = int(ENV['fab_port'])
+    def_port = ENV['fab_port']
     username = ENV['fab_user']
     password = ENV['fab_password']
     pkey = ENV['fab_pkey']
     key_filename = ENV['fab_key_filename']
     for host in ENV['fab_hosts']:
+        host, _, port = host.partition(':')
+        portnr = int(port or def_port)
         client = ssh.SSHClient()
         client.load_system_host_keys()
         if 'fab_new_host_key' in ENV and ENV['fab_new_host_key'] == 'accept':
             client.set_missing_host_key_policy(ssh.AutoAddPolicy())
         client.connect(
-            host, port, username, password, pkey, key_filename
+            host, portnr, username, password, pkey, key_filename
         )
         CONNECTIONS.append((host, client))
     if not CONNECTIONS:
         print("The fab_hosts list was empty.")
         print("Please specify some hosts to connect to.")
-        exit(1)
+        sys.exit(1)
 
 def _disconnect():
     "Disconnect all clients."
@@ -755,6 +895,9 @@ def _disconnect():
 
 def _lazy_format(string, env=ENV):
     "Do recursive string substitution of ENV vars - both lazy and earger."
+    if string is None:
+        return None
+    env = dict([(k, str(v)) for k, v in env.items()])
     def replacer_fn(match):
         var = match.group('var')
         if var in env:
@@ -764,74 +907,38 @@ def _lazy_format(string, env=ENV):
     return re.sub(_LAZY_FORMAT_SUBSTITUTER, replacer_fn, string % env)
 
 def _on_hosts_do(fn, *args, **kvargs):
-    """Invoke the given function with hostname and client parameters in
+    """
+    Invoke the given function with hostname and client parameters in
     accord with the current fac_mode strategy.
     
-    fn should be of type:
-        (str:hostname, paramiko.SSHClient:clinet) -> bool:success
+    fn should be a callable taking these parameters:
+        hostname : str
+        client : paramiko.SSHClient
+        *args
+        **kvargs
     
     """
     strategy = ENV['fab_mode']
-    if strategy == 'fanout':
-        threads = []
-        for host, client in CONNECTIONS:
-            env = dict(ENV)
-            env['fab_host'] = host
-            functor = lambda: fn(host, client, env, *args, **kvargs)
-            thread = threading.Thread(None, functor)
-            thread.setDaemon(True)
-            thread.start()
-            threads.append(thread)
-        for thread in threads:
-            thread.join()
-    elif strategy == 'rolling':
-        for host, client in CONNECTIONS:
-            env = dict(ENV)
-            env['fab_host'] = host
-            fn(host, client, env, *args)
+    if strategy in STRATEGIES:
+        strategy_fn = STRATEGIES[strategy]
+        successful = strategy_fn(fn, *args, **kvargs)
+        if not successful:
+            print("Operation failed: " + fn.__name__)
+            sys.exit(1)
     else:
         print("Unsupported fab_mode: %s" % strategy)
-        print("Supported modes are: fanout, rolling")
-        exit(1)
+        print("Supported modes are: %s" % (', '.join(STRATEGIES.keys())))
+        sys.exit(1)
 
-def _put(host, client, env, localpath, remotepath, **kvargs):
-    ftp = client.open_sftp()
-    localpath = _lazy_format(localpath, env)
-    remotepath = _lazy_format(remotepath, env)
-    print("[%s] put: %s -> %s" % (host, localpath, remotepath))
-    ftp.put(localpath, remotepath)
+def _confirm_proceed(exec_type, host, kvargs):
+    if 'confirm' in kvargs:
+        infotuple = (exec_type, host, _lazy_format(kvargs['confirm']))
+        question = "Confirm %s for host %s: %s [yN] " % infotuple
+        answer = raw_input(question)
+        return answer in 'yY'
+    return True
 
-def _download(host, client, env, remotepath, localpath, **kvargs):
-    ftp = client.open_sftp()
-    localpath = _lazy_format(localpath) + '.' + host
-    remotepath = _lazy_format(remotepath)
-    print("[%s] download: %s <- %s" % (host, localpath, remotepath))
-    ftp.get(remotepath, localpath)
-
-def _run(host, client, env, cmd, **kvargs):
-    cmd = _lazy_format(cmd, env)
-    real_cmd = env['fab_shell'] % cmd.replace('"', '\\"')
-    print("[%s] run: %s" % (host, (env['fab_debug'] and real_cmd or cmd)))
-    stdin, stdout, stderr = client.exec_command(real_cmd)
-    out_th = _start_outputter("[%s] out" % host, stdout)
-    err_th = _start_outputter("[%s] err" % host, stderr)
-    out_th.join()
-    err_th.join()
-
-def _sudo(host, client, env, cmd, **kvargs):
-    cmd = _lazy_format(cmd, env)
-    real_cmd = env['fab_shell'] % ("sudo -S " + cmd.replace('"', '\\"'))
-    print("[%s] sudo: %s" % (host, (env['fab_debug'] and real_cmd or cmd)))
-    stdin, stdout, stderr = client.exec_command(real_cmd)
-    stdin.write(env['fab_password'])
-    stdin.write('\n')
-    stdin.flush()
-    out_th = _start_outputter("[%s] out" % host, stdout)
-    err_th = _start_outputter("[%s] err" % host, stderr)
-    out_th.join()
-    err_th.join()
-
-def _get_failcode(kvarg_map):
+def _get_failcode(kvarg_map, default='abort'):
     codes = {
         'ignore': 1,
         'warn': 2,
@@ -840,7 +947,7 @@ def _get_failcode(kvarg_map):
     if 'fail' in kvarg_map:
         return codes[kvarg_map['fail']]
     else:
-        return codes['abort']
+        return codes[default]
 
 def _start_outputter(prefix, channel):
     def outputter():
@@ -862,42 +969,56 @@ def _pick_fabfile():
     else:
         return guesses[0] # load() will barf for us...
 
-def _validate_commands(args):
+def _load_default_settings():
+    "Load user-default fabric settings from ~/.fabric"
+    # TODO: http://mail.python.org/pipermail/python-list/2006-July/393819.html
+    cfg = os.path.expanduser("~/.fabric")
+    if os.path.exists(cfg):
+        comments = lambda s: s and not s.startswith("#")
+        settings = filter(comments, open(cfg, 'r'))
+        settings = [(k.strip(),v.strip()) for k,_,v in
+            [s.partition('=') for s in settings]]
+        ENV.update(settings)
+
+def _parse_args(args):
+    cmds = []
     for cmd in args:
-        if cmd.find(':') != -1:
-            cmd = cmd.split(':', 1)[0]
-        if not cmd in COMMANDS:
-            print("No such command: %s" % cmd)
-            _list_commands()
-            exit(1)
-    if not args:
+        cmd_args = {}
+        if ':' in cmd:
+            cmd, cmd_str_args = cmd.split(':', 1)
+            for cmd_arg_kv in cmd_str_args.split(','):
+                k, _, v = cmd_arg_kv.partition('=')
+                cmd_args[k] = (v % ENV)
+        cmds.append((cmd, cmd_args))
+    return cmds
+
+def _validate_commands(cmds):
+    if not cmds:
         print("No commands given.")
         _list_commands()
+    else:
+        for cmd in cmds:
+            if not cmd[0] in COMMANDS:
+                print("No such command: %s" % cmd[0])
+                sys.exit(1)
 
-def _execute_commands(args):
-    for cmd in args:
-        cmd_name = cmd
-        cmd_args = None
-        if cmd.find(':') != -1:
-            cmd_name, cmd_args = cmd.split(':', 1)
-        ENV['fab_cur_command'] = cmd_name
-        print("Running %s..." % cmd_name)
-        if cmd_args is not None:
-            cmd_arg_kvs = {}
-            for cmd_arg_kv in cmd_args.split(','):
-                k, _, v = cmd_arg_kv.partition('=')
-                cmd_arg_kvs[k] = (v % ENV)
-            COMMANDS[cmd_name](**cmd_arg_kvs)
-        else:
-            COMMANDS[cmd]()
+def _execute_commands(cmds):
+    for cmd, args in cmds:
+        ENV['fab_cur_command'] = cmd
+        print("Running %s..." % cmd)
+        if args is not None:
+            args = dict(zip(args.keys(), map(_lazy_format, args.values())))
+        COMMANDS[cmd](**(args or {}))
 
 def main(args):
     try:
         print(__greeter__ % ENV)
         fabfile = _pick_fabfile()
-        load(fabfile)
-        _validate_commands(args)
-        _execute_commands(args)
+        _load_default_settings()
+        load(fabfile, fail='warn')
+        commands = _parse_args(args)
+        _validate_commands(commands)
+        _execute_commands(commands)
     finally:
         _disconnect()
         print("Done.")
